"""

This library is an API wrapper for Yandex Delivery

Author: Egor Udalov
Latest version: 1.0.1

Quickstart guide: https://yandex.ru/dev/logistics/api/about/quickstart.html

"""

from .delivery_api import YandexDeliveryApi
from . import base_methods



"""

This module describes API objects for API Base Methods

"""
from typing import Optional, Literal, List, Tuple
from typing_extensions import Annotated

from pydantic import (BaseModel, PositiveInt, NonNegativeFloat,
                      Field, HttpUrl, NonNegativeInt,
                      StrictBool, EmailStr, field_validator)
from pydantic_extra_types.phone_numbers import PhoneNumber


class Customer(BaseModel):
    """

    This class describes customer. It's useful for on delivery payment

    """
    email: Optional[EmailStr] = None
    inn:   Optional[str] = None
    phone: Optional[Annotated[str, PhoneNumber]] = None


class PaymentOnDelivery(BaseModel):
    """

    This class describes on delivery payment info

    """
    customer:       Optional[Customer] = None
    payment_method: Literal["card", "cash"]


class ExternalOrderCost(BaseModel):
    """

    This class describes external order cost

    """
    currency:      Annotated[str, Field(min_length=1)]
    currency_sign: Annotated[str, Field(min_length=1)]
    value:         Annotated[str, Field(min_length=1)]


class Contact(BaseModel):
    """

    This class describes contact person

    """
    email:                 Optional[EmailStr] = None
    name:                  Annotated[str, Field(min_length=1)]
    phone:                 Annotated[str, PhoneNumber]
    phone_additional_code: Optional[str] = None


class Buyout(BaseModel):
    """

    This class describes chosen buyout type

    """
    payment_method: Literal["card", "cash"]


class Address(BaseModel):
    """

    This class describes address fields

    """
    building:        Optional[str] = None
    building_name:   Optional[str] = None
    city:            Optional[str] = None
    comment:         Optional[str] = None
    coordinates:     Tuple[float, float]
    country:         Optional[str] = None
    description:     Optional[str] = None
    door_code:       Optional[str] = None
    door_code_extra: Optional[str] = None
    doorbell_name:   Optional[str] = None
    fullname:        Annotated[str, Field(min_length=1)]
    porch:           Optional[str] = None
    sflat:           Optional[str] = None
    sfloor:          Optional[str] = None
    shortname:       Optional[str] = None
    street:          Optional[str] = None
    uri:             Optional[str] = None


class RoutePoint(BaseModel):
    """

    This class describes route point parameters

    """
    address:             Address
    buyout:              Optional[Buyout] = None
    contact:             Contact
    external_order_cost: Optional[ExternalOrderCost] = None
    external_order_id:   Optional[str] = None
    leave_under_door:    Optional[StrictBool] = None
    meet_outside:        Optional[StrictBool] = None
    no_door_call:        Optional[StrictBool] = None
    payment_on_delivery: Optional[PaymentOnDelivery] = None
    pickup_code:         Optional[Annotated[str, Field(min_length=6, max_length=6)]] = None
    point_id:            int  # int64
    skip_confirmation:   Optional[StrictBool] = None
    type:                Literal["source", "destination", "return"]
    visit_order:         PositiveInt


class EmergencyContact(BaseModel):
    """

    This class describes delivery customer contact

    """
    name:                  Annotated[str, Field(min_length=1)]
    phone:                 Annotated[str, PhoneNumber]
    phone_additional_code: Optional[str] = None


class ClientRequirements(BaseModel):
    """

    This class describes client requirements, these requirements are
    necessary to create or edit claim

    """
    assign_robot:  Optional[StrictBool] = None
    cargo_loaders: Optional[Literal[0, 1, 2]] = None
    cargo_options: Optional[List[Literal["thermobag", "auto_courier"]]] = None
    cargo_type:    Optional[Literal["van", "lcv_m", "lcv_l"]] = None
    pro_courier:   Optional[StrictBool] = None
    taxi_class:    Literal["courier", "express", "cargo"]


class CallbackProperties(BaseModel):
    """

    This class describes callback URL for API to send claim status

    """
    callback_url: Optional[HttpUrl] = None


class ItemSize(BaseModel):
    """

    This class describes item size in meters

    """
    height: NonNegativeFloat
    length: NonNegativeFloat
    width:  NonNegativeFloat


class Item(BaseModel):
    """

    This class describes item to delivery
    All integer values should be int64

    """
    title:         Annotated[str, Field(min_length=1)]
    quantity:      PositiveInt
    cost_value:    Annotated[str | float, Field(ge=0)]
    size:          ItemSize
    droppof_point: NonNegativeInt
    pickup_point:  NonNegativeInt
    cost_currency: Annotated[str, Field(default="RUB", min_length=1)] = "RUB"
    extra_id:      Optional[str] = None
    fiscalization: Optional[str] = None
    weight:        Optional[NonNegativeFloat] = None

    @field_validator("cost_value", mode="before")
    def convert_to_float(cls, value):
        """
        Any to float
        """
        if not isinstance(value, float):
            return float(value)
        return value

    @field_validator("cost_value", mode="after")
    def convert_to_string(cls, value):
        """
        Any to string
        """
        return str(value)


class Claim(BaseModel):
    """

    This class describes delivery claim

    """
    auto_accept:           Optional[StrictBool] = None
    callback_properties:   Optional[CallbackProperties] = None
    client_requirements:   Optional[ClientRequirements] = None
    comment:               Optional[str] = None
    due:                   None = None  # TODO
    emergency_contact:     Optional[EmergencyContact] = None
    items:                 Optional[List[Item]] = None
    offer_payload:         Optional[str] = None
    optional_return:       Optional[StrictBool] = None
    referral_source:       Optional[str] = None
    route_points:          List[RoutePoint]
    same_day_data:         None = None  # TODO
    shipping_document:     Optional[str] = None
    skip_act:              Optional[StrictBool] = None
    skip_client_notify:    Optional[StrictBool] = None
    skip_door_to_door:     Optional[StrictBool] = None
    skip_emergency_notify: Optional[StrictBool] = None



"""

This module describes API class26

"""

import logging
import uuid

from typing import Literal

from .base_methods import Claim
from .rest_adapter import RestAdapter
from .rest_result import Result


class YandexDeliveryApi:
    """

    This class describes Delivery API object

    """
    def __init__(self,
                 hostname:        str = "b2b.taxi.yandex.net/b2b/cargo/integration",
                 api_key:         str = "",
                 ver:             str = "v2",
                 content_type:    str = "application/json",
                 accept_language: str = "ru",
                 retries:         int = 0,
                 timeout:         int | None = 5,
                 logger:          logging.Logger = None):
        self._rest_adapter = RestAdapter(hostname=hostname,
                                         api_key=api_key,
                                         ver=ver,
                                         content_type=content_type,
                                         accept_language=accept_language,
                                         retries=retries,
                                         timeout=timeout,
                                         logger=logger)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def create(self,
                     claim: Claim) -> Result:
        """
        Request to /claims/create method

        :param claim: Claim object

        https://yandex.ru/dev/logistics/api/ref/basic/IntegrationV2ClaimsCreate.html
        """
        return await self._rest_adapter.post(endpoint="/claims/create",
                                             payload=claim.model_dump(mode="json",
                                                                      exclude_none=True),
                                             params={"request_id": str(uuid.uuid4())})

    async def check_price(self,
                          claim: Claim) -> Result:
        """
        Request to /check-price method

        :param claim: Claim object

        https://yandex.ru/dev/logistics/api/ref/estimate/IntegrationV2CheckPrice.html
        """
        return await self._rest_adapter.post(endpoint="/check-price",
                                             payload=claim.model_dump(mode="json",
                                                                      exclude_none=True))

    async def accept(self,
                     claim_id: str,
                     version: int = 1) -> Result:
        """
        Request to /claims/accept

        :param claim_id: Claim ID
        :param version: Claim version. 1 by default (claim was not edited)

        https://yandex.ru/dev/logistics/api/ref/basic/IntegrationV2ClaimsAccept.html
        """
        return await self._rest_adapter.post(endpoint="/claims/accept",
                                             params={"claim_id": claim_id},
                                             payload={"version": version})

    async def info(self,
                   claim_id: str) -> Result:
        """
        Request to /claims/info method

        :param claim_id: Claim ID

        https://yandex.ru/dev/logistics/api/ref/basic/IntegrationV2ClaimsInfo.html
        """
        return await self._rest_adapter.post(endpoint="/claims/info",
                                             params={"claim_id": claim_id})

    async def cancel_info(self,
                          claim_id: str) -> Result:
        """
        Request to /claims/cancel-info method

        :param claim_id: Claim ID

        https://yandex.ru/dev/logistics/api/ref/cancel-and-skip-points/IntegrationV2ClaimsCancelInfo.html
        """
        return await self._rest_adapter.post(endpoint="/claims/cancel-info",
                                             params={"claim_id": claim_id})

    async def cancel(self,
                     claim_id: str,
                     cancel_state: Literal["free", "paid"],
                     version: int = 1) -> Result:
        """
        Request to /claims/cancel method

        :param claim_id: Claim ID
        :param cancel_state: Cancelling status. Enum: ["free", "paid"]
        :param version: Claim version. 1 by default (claim was not edited)

        https://yandex.ru/dev/logistics/api/ref/cancel-and-skip-points/IntegrationV2ClaimsCancelInfo.html
        """
        return await self._rest_adapter.post(endpoint="/claims/cancel",
                                             params={"claim_id": claim_id},
                                             payload={"cancel_state": cancel_state,
                                                      "version": version})

    async def driver_voiceforwarding(self,
                                     claim_id: str,
                                     point_id: int = None) -> Result:
        """
        Request to /driver_voiceforwarding method

        :param claim_id: Claim ID
        :param point_id: Point ID (generated by Delivery)

        https://yandex.ru/dev/logistics/api/ref/performer-info/IntegrationV2DriverVoiceForwarding.html
        """
        return await self._rest_adapter.post(endpoint="/driver-voiceforwarding",
                                             payload={"claim_id": claim_id,
                                                      "point_id": point_id} if point_id
                                             else {"claim_id": claim_id})

    async def performer_position(self,
                                 claim_id: str) -> Result:
        """
        Request to /performer_position method

        :param claim_id: Claim ID

        https://yandex.ru/dev/logistics/api/ref/performer-info/IntegrationV2ClaimsPerformerPosition.html
        """
        return await self._rest_adapter.get(endpoint="/claims/performer-position",
                                            params={"claim_id": claim_id})

    async def tracking_links(self,
                             claim_id: str) -> Result:
        """
        Request to /tracking-links method

        :param claim_id: Claim ID

        https://yandex.ru/dev/logistics/api/ref/performer-info/IntegrationV2ClaimsTrackingLinks.html
        """
        return await self._rest_adapter.get(endpoint="/claims/tracking-links",
                                            params={"claim_id": claim_id})



"""
This module describes REST adapter result class
"""

from types import SimpleNamespace
from dataclasses import dataclass

from aiohttp.typedefs import CIMultiDictProxy


@dataclass
class Result:
    """
    REST adapter result class
    """
    def __init__(self,
                 status_code: int,
                 headers:     CIMultiDictProxy[str],
                 message:     str = '',
                 data:        dict = None):
        """
        Result returned from low-level RestAdapter
        :param status_code: Standard HTTP Status code
        :param message: Human readable result
        :param data: namespace object
        """
        self.status_code = int(status_code)
        self.headers = headers
        self.message = str(message)
        self.json = data
        self.data = SimpleNamespace(**data) if data else {}



"""

This module describes REST adapter class

"""

import logging
from json import JSONDecodeError

import aiohttp

from .exceptions import YandexDeliveryApiError
from .rest_result import Result


class RestAdapter:
    """

    This class describes common REST operations

    """
    def __init__(self,
                 hostname:        str,
                 api_key:         str,
                 ver:             str,
                 content_type:    str,
                 accept_language: str,
                 timeout:         int = None,
                 retries:         int = 0,
                 logger:          logging.Logger = None):
        """

        RestAdapter Constructor

        :param hostname: b2b.taxi.yandex.net/b2b/cargo/integration
        :param api_key: OAuth-token
        :param ver: API version
        :param content_type: Content-Type header
        :param accept_language: Accept-Language header
        :param timeout: Timeout in seconds / None to request without timeout
        :param retries: Additional attempts
        :param logger: (optional) logger instance
        """
        self._logger = logger or logging.getLogger(__package__)
        self.url = f"https://{hostname}/{ver}"
        self.content_type = content_type
        self.accept_language = accept_language
        self._api_key = api_key
        self._timeout = aiohttp.ClientTimeout(total=timeout)
        self._retries = retries

    async def _do(self,
                  http_method: str,
                  endpoint:    str,
                  params:      dict[str, str] = None,
                  payload:     dict = None) -> Result:
        """

        REST operations common method

        :param http_method: GET, POST, DELETE, etc.
        :param endpoint: URL Endpoint
        :param params: Params
        :param payload: Dictionary with payload
        :return: Result instance
        """
        full_url = self.url + endpoint
        headers = {'content-type': self.content_type, 'Accept-Language': self.accept_language,
                   'Authorization': f'Bearer {self._api_key}'}
        log_line_pre = f"method={http_method}, url={full_url}, params={params}"
        log_line_post = ', '.join((log_line_pre, "success={}, status_code={}, message={}"))

        for _ in range(1 + self._retries):
            try:
                self._logger.debug(msg=log_line_pre)
                async with aiohttp.ClientSession(headers=headers, timeout=self._timeout) as session:
                    response = await session.request(
                        method=http_method,
                        url=self.url + endpoint,
                        params=params,
                        json=payload
                    )

                    try:
                        data_out = await response.json()
                    except (ValueError, TypeError, JSONDecodeError) as e:
                        self._logger.error(msg=log_line_post.format(False, None, e))
                        raise YandexDeliveryApiError("Bad JSON in response") from e

                break
            except aiohttp.ClientError as e:
                self._logger.error(msg=str(e))
                raise YandexDeliveryApiError("Request failed") from e
            except (TimeoutError, aiohttp.ServerTimeoutError) as e:
                self._logger.error(msg=str(e))
                if _ >= self._retries:
                    raise YandexDeliveryApiError("Timeout error") from e
                continue

        is_success = 299 >= response.status >= 200
        if is_success:
            return Result(response.status,
                          headers=response.headers,
                          message=response.reason,
                          data=data_out)
        raise YandexDeliveryApiError(f"{response.status}: {response.reason} "
                                     f"({data_out['message']})")

    async def get(self,
                  endpoint: str,
                  params: dict[str, str] = None,
                  payload: dict = None) -> Result:
        """
        Implements GET method
        """
        return await self._do(http_method='GET',
                              endpoint=endpoint,
                              params=params,
                              payload=payload)

    async def post(self,
                   endpoint: str,
                   params: dict[str, str] = None,
                   payload: dict = None) -> Result:
        """
        Implements POST method
        """
        return await self._do(http_method='POST',
                              endpoint=endpoint,
                              params=params,
                              payload=payload)

    async def delete(self,
                     endpoint: str,
                     params: dict[str, str] = None,
                     payload: dict = None) -> Result:
        """
        Implements DELETE method
        """
        return await self._do(http_method='DELETE',
                              endpoint=endpoint,
                              params=params,
                              payload=payload)



from setuptools import setup

setup(name='yandex-delivery-api',
      version='1.0.3',
      description='Python wrapper for Yandex Delivery API',
      packages=['yandex_delivery'],
      author_email='edm.rud3nko@yandex.ru',
      zip_safe=False)
